<?php

namespace Drp\SimpleCms;

use Drp\SimpleCms\Contracts\Publishable;
use Drp\SimpleCms\Events\ContentForReviewWasRejected;
use Drp\SimpleCms\Events\ContentWasPublished;
use Drp\SimpleCms\Events\ContentWasSentForReview;
use Drp\SimpleCms\Events\ContentWasUnPublished;
use Drp\SimpleCms\Exceptions\CouldNotResetDraft;
use Drp\SimpleCms\Exceptions\InvalidContentStatus;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;

trait Publisher
{
    public function generateContentId(): int
    {
        $last = $this->orderByDesc('content_id')->first();

        return is_null($last) ? 1 : $last->content_id + 1;
    }

    public function save(array $options = []): bool
    {
        $this->content_id = $this->content_id ?: $this->generateContentId();

        $this->ensureStatusIntegrity();

        return parent::save($options);
    }

    private function ensureStatusIntegrity(): void
    {
        $existing = $this->versionAlreadyExists($this->id);

        if ($existing) {
            throw new InvalidContentStatus('Content already exists with this content_id and status');
        }
    }

    private function versionAlreadyExists(?int $currentId = null): ?Publishable
    {
        return $this
            ->where('content_id', $this->content_id)
            ->where('status', $this->status)
            ->where($this->getQualifiedKeyName(), '!=', $currentId)
            ->first();
    }

    public function getPublishedVersion(): ?Publishable
    {
        if ($this->isPublished()) {
            return $this;
        }

        return $this
            ->where('content_id', $this->content_id)
            ->wherePublished()
            ->first();
    }

    public function getUnpublishedVersion(): ?Publishable
    {
        if ($this->isUnPublished()) {
            return $this;
        }

        return $this
            ->where('content_id', $this->content_id)
            ->WhereUnPublished()
            ->first();
    }

    public function getDraftOrReviewVersion(): ?Publishable
    {
        if (!$this->isPublished() && !$this->isUnPublished()) {
            return $this;
        }

        return $this
            ->where('content_id', $this->content_id)
            ->whereDraftOrInReview()
            ->first();
    }

    public function scopeWherePublished(Builder $query): Builder
    {
        return $query->where('status', Publishable::STATUS_PUBLISHED);
    }

    public function scopeWhereUnPublished(Builder $query): Builder
    {
        return $query->where('status', Publishable::STATUS_UNPUBLISHED);
    }

    public function scopeWhereDraftOrInReview(Builder $query): Builder
    {
        return $query->where(function ($query): void {
            $query->where('status', '=', Publishable::STATUS_DRAFT);
            $query->orWhere('status', '=', Publishable::STATUS_REVIEW);
        });
    }

    public function scopeWhereDraft(Builder $query): Builder
    {
        return $query->where('status', '=', Publishable::STATUS_DRAFT);
    }

    public function scopeWhereInReview(Builder $query): Builder
    {
        return $query->where('status', '=', Publishable::STATUS_REVIEW);
    }

    public function isPublished(): bool
    {
        return $this->status === Publishable::STATUS_PUBLISHED;
    }

    public function isUnPublished(): bool
    {
        return $this->status === Publishable::STATUS_UNPUBLISHED;
    }

    public function isDraft(): bool
    {
        return $this->status === Publishable::STATUS_DRAFT;
    }

    public function isInReview(): bool
    {
        return $this->status === Publishable::STATUS_REVIEW;
    }

    public function isDraftOrInReview(): bool
    {
        return $this->isDraft() || $this->isInReview();
    }

    public function hasPublishedVersion(): bool
    {
        return !empty($this->getPublishedVersion());
    }

    public function hasDraftOrReviewVersion(): bool
    {
        return !empty($this->getDraftOrReviewVersion());
    }

    public function hasUnPublishedVersion(): bool
    {
        return !empty($this->getUnpublishedVersion());
    }

    public function canBePublished(): bool
    {
        return $this->exists;
    }

    public function canBeReviewed(): bool
    {
        return $this->isDraftOrInReview();
    }

    public function canBeReset(): bool
    {
        return $this->hasDraftOrReviewVersion() && $this->hasPublishedVersion();
    }

    public function updatePublishStatus(): void
    {
        $this->isPublished() ? $this->unPublish() : $this->publish();
    }

    public function updateReviewStatus(): void
    {
        $this->isInReview() ? $this->markAsDraft() : $this->markAsInReview();
    }

    private function markAsInReview(): void
    {
        $this->status = Publishable::STATUS_REVIEW;
        $this->save();

        event(new ContentWasSentForReview($this));
    }

    private function markAsDraft(): void
    {
        $this->status = Publishable::STATUS_DRAFT;
        $this->save();

        event(new ContentForReviewWasRejected($this));
    }

    private function publish(): void
    {
        if ($this->isDraftOrInReview()) {
            $this->removeExistingVersion();
        }

        $this->status = Publishable::STATUS_PUBLISHED;
        $this->save();

        event(new ContentWasPublished($this));
    }

    private function unPublish(): void
    {
        $this->status = Publishable::STATUS_UNPUBLISHED;
        $this->save();

        event(new ContentWasUnPublished($this));
    }

    private function removeExistingVersion(): void
    {
        // phpcs:disable SlevomatCodingStandard.ControlStructures.EarlyExit.EarlyExitNotUsed
        $published = $this->getPublishedVersion();
        if ($published) {
            $published->delete();
        }

        $unPublished = $this->getUnpublishedVersion();
        if ($unPublished) {
            $unPublished->delete();
        }
        // phpcs:enable SlevomatCodingStandard.ControlStructures.EarlyExit.EarlyExitNotUsed
    }

    public function resetDraft(): void
    {
        if (!$this->canBeReset()) {
            throw new CouldNotResetDraft();
        }

        $draft = $this->getDraftOrReviewVersion();
        $draft->delete();
    }

    public static function getLatestVersionById(int $contentId): Content
    {
        return self::where('content_id', $contentId)
            ->orderBy('status')
            ->firstOrFail();
    }

    public static function getAllVersionsPrioritisedByStatus(?array $with = []): ?Collection
    {
        return self::with($with)
            ->get()
            ->groupBy('content_id')
            ->map(function ($content) {
                $keyed = $content->keyBy('status');
                //Order by published status
                return $keyed->get(
                    Publishable::STATUS_PUBLISHED,
                    $keyed->get(
                        Publishable::STATUS_UNPUBLISHED,
                        $keyed->get(
                            Publishable::STATUS_REVIEW,
                            $keyed->get(Publishable::STATUS_DRAFT)
                        )
                    )
                );
            })
            ->flatten();
    }

    /**
     * Return the options for the select box
     * using getAllPublishedOrFirstDraft.
     *
     * @return mixed
     */
    public static function optionsForSelect()
    {
        return self::getAllVersionsPrioritisedByStatus()
            ->pluck('title', 'content_id')
            ->toArray();
    }

    public function beforeSave(): ?\Closure
    {
        return null;
    }

    public function afterSave(): ?\Closure
    {
        return null;
    }

    public function isContentDirty(array $attributes): bool
    {
        if (!$this->exists()) {
            return true;
        }

        return $this->fresh()->fill($attributes)->isDirty();
    }
}
